package zero1hd.rhythmbullet.audio;

import java.util.Observable;
import java.util.Observer;
import java.util.Random;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Preferences;
import com.badlogic.gdx.audio.Music;
import com.badlogic.gdx.audio.Music.OnCompletionListener;
import com.badlogic.gdx.files.FileHandle;

/**
 * Manages current games music playback and does this in tandem with the {@link MusicList} by asking to retrieve files and then feeding it to LibGDX.
 * Notifies observers when a new song is loaded.
 * The loading model is like taking a disk and loading it into a player. It doesn't necessarily mean it'll play right away, but its ready and the only track that has a stream opened.
 * @author yunya
 *
 */
public class MusicController extends Observable implements OnCompletionListener, Observer {
	public enum States {
		Loaded, Playing;
	}
	
	private MusicList musicList;
	private MinimalAudioHeader musicHeader;
	private Music music;
	private int currentPlaybackIndex;
	private boolean autoPlay;
	private boolean shuffle;
	private Random rand;
	private Preferences prefs;
	
	public MusicController(MusicList musicList, Preferences prefs) {
		if (prefs == null) throw new NullPointerException("preferences can't be null...");
		if (musicList == null) throw new NullPointerException("music list can't be null...");
		musicList.addObserver(this);
		this.prefs = prefs;
		this.musicList = musicList;
		rand = new Random();
	}
	
	/**
	 * This play method automatically sets the volume.
	 */
	public void play() {
		if (music != null) {
			Gdx.app.debug("MusicListController", "Playing from MLC.");
			music.play();
			music.setVolume(prefs.getFloat("music vol", 1f));
			notifyObservers(States.Playing);
		} else {
			Gdx.app.debug("MusicListController", "failed to begin playing. Load the music!!!");
		}
	}
	
	/**
	 * Called to pause current song. Does nothing if no song is playing or loaded.
	 */
	public void pause() {
		if (music != null) {
			music.pause();
		}
	}
	
	/**
	 * Loads music based on the index in the {@link MusicList}.
	 * @param index of music to play
	 */
	public void setMusicByIndex(int index) {
		this.currentPlaybackIndex = index;
		loadMusic();
	}
	
	/**
	 * Goes to the next track
	 */
	public void skip() {
		currentPlaybackIndex++;
		if (shuffle) {
			shuffle(false);
		}
		loadMusic();
	}
	
	/**
	 * Goes to the previous track
	 */
	public void previous() {
		currentPlaybackIndex--;
		if (shuffle) {
			shuffle(false);
		}
		loadMusic();
	}

	@Override
	public void onCompletion(Music music) {
		if (autoPlay) {
			if (shuffle) {
				shuffle(false);
			} else {
				currentPlaybackIndex++;
			}
			loadMusic();
			play();
		}
	}
	
	/**
	 * Shuffles the controller whether the shuffle boolean is true or false.
	 * @param load whether this method should also make sure to load the music, dispose of the last one, etc. Normally called unless you plan to manually call it elsewhere.
	 */
	public void shuffle(boolean load) {
		Gdx.app.debug("MusicListController", "shuffled.");
		if (musicList.getTotal() == 0) {
			currentPlaybackIndex = 0;
		} else {
			currentPlaybackIndex = rand.nextInt(musicList.getTotal());
		}
		if (load) {
			loadMusic();
		}
	}
	
	public void setAutoPlay(boolean autoPlay) {
		this.autoPlay = autoPlay;
	}
	
	public void setShuffle(boolean shuffle) {
		this.shuffle = shuffle;
	}
	
	public boolean isShuffle() {
		return shuffle;
	}
	
	public boolean isAutoPlay() {
		return autoPlay;
	}
	
	/**
	 * Loads the current selected song.
	 */
	public void loadMusic() {
		Gdx.app.debug("MusicListController", "music is being loaded and listeners are being notified.");
		musicHeader = null;
		if (music != null) {
			music.dispose();
		}
		if (currentPlaybackIndex < 0) {
			currentPlaybackIndex = musicList.getTotal()-1;
		}
		if (currentPlaybackIndex >= musicList.getTotal()) {
			currentPlaybackIndex = 0;
		}
		this.music = Gdx.audio.newMusic(musicList.getMusicArray().get(currentPlaybackIndex));
		music.setOnCompletionListener(this);
		
		setChanged();
		
		if (autoPlay) {
			play();
		}
		notifyObservers(States.Loaded);
	}
	
	public MusicList getMusicList() {
		return musicList;
	}
	
	public FileHandle getCurrentMusicFileHandle() {
		return musicList.getSongFileHandleFromIndex(currentPlaybackIndex);
	}
	
	public MinimalAudioHeader getCurrentMusicHeader() {
		if (musicHeader != null) {
			return musicHeader;
		} else {
			return musicHeader = musicList.newMinimalAudioHeader(getCurrentMusicFileHandle());
		}
	}
	
	@Override
	public void update(Observable o, Object arg) {
		if (o == musicList) {
			loadMusic();
			if (autoPlay) {
				play();
			}
		}
	}
	
	public String getCurrentSongName() {
		return getCurrentMusicFileHandle().nameWithoutExtension();
	}
	
	public float getCurrentPosition() {
		if (music != null) {
			return music.getPosition();
		}
		return 0;
	}
	
	public void setMusicPosition(float position) {
		if (music != null) {
			music.setPosition(position);
		}
	}
	
	public boolean isPlaying() {
		return music.isPlaying();
	}
	
	/**
	 * Returns the current music. In no circumstances should this be used to begin playing or it would be playing independent of the controller.
	 * Doing otherwise would mean you are playing the music separately of the controller.
	 * @return
	 */
	public Music getCurrentMusic() {
		return music;
	}
}
